@@ -21,7 +21,8 @@ from enum import IntEnum
 
 from hypothesis.errors import Frozen, InvalidArgument
 from hypothesis.internal.compat import hbytes, hrange, text_type, \
-    int_to_bytes, benchmark_time, unicode_safe_repr
+    bit_length, int_to_bytes, benchmark_time, int_from_bytes, \
+    unicode_safe_repr
 
 
 def uniform(random, n):
@@ -45,13 +46,14 @@ class StopTest(BaseException):
 global_test_counter = 0
 
 
-MAX_DEPTH = 50
+MAX_DEPTH = 100
 
 
 class ConjectureData(object):
 
     @classmethod
     def for_buffer(self, buffer):
+        buffer = hbytes(buffer)
         return ConjectureData(
             max_length=len(buffer),
             draw_bytes=lambda data, n, distribution:
@@ -78,7 +80,8 @@ class ConjectureData(object):
         global_test_counter += 1
         self.start_time = benchmark_time()
         self.events = set()
-        self.bind_points = set()
+        self.forced_indices = set()
+        self.capped_indices = {}
 
     def __assert_not_frozen(self, name):
         if self.frozen:
@@ -116,19 +119,6 @@ class ConjectureData(object):
             if not self.frozen:
                 self.stop_example()
 
-    def mark_bind(self):
-        """Marks a point as somewhere that a bind occurs - that is, data
-        drawn after this point may depend significantly on data drawn prior
-        to this point.
-
-        Having points like this explicitly marked allows for better shrinking,
-        as we run a pass which tries to shrink the byte stream prior to a bind
-        point while rearranging what comes after somewhat to allow for more
-        flexibility. Trying that at every point in the data stream is
-        prohibitively expensive, but trying it at a couple dozen is basically
-        fine."""
-        self.bind_points.add(self.index)
-
     def start_example(self):
         self.__assert_not_frozen('start_example')
         self.interval_stack.append(self.index)
@@ -171,35 +161,60 @@ class ConjectureData(object):
     def draw_bytes(self, n):
         return self.draw_bytes_internal(n, uniform)
 
-    def write(self, string):
-        assert isinstance(string, hbytes)
-
-        def distribution(random, n):
-            assert n == len(string)
-            return string
-
-        while True:
-            x = self.draw_bytes_internal(len(string), distribution)
-            if x == string:
-                return
-
-    def draw_bytes_internal(self, n, distribution):
+    def draw_bits(self, n):
+        self.__assert_not_frozen('draw_bits')
         if n == 0:
-            return hbytes(b'')
-        self.__assert_not_frozen('draw_bytes')
-        initial = self.index
+            result = 0
+        elif n % 8 == 0:
+            return int_from_bytes(self.draw_bytes(n // 8))
+        else:
+            n_bytes = (n // 8) + 1
+            self.__check_capacity(n_bytes)
+            buf = bytearray(self._draw_bytes(self, n_bytes, uniform))
+            assert len(buf) == n_bytes
+            mask = (1 << (n % 8)) - 1
+            buf[0] &= mask
+            self.capped_indices[self.index] = mask
+            buf = hbytes(buf)
+            self.__write(buf)
+            result = int_from_bytes(buf)
+        assert bit_length(result) <= n
+        return result
+
+    def write(self, string):
+        self.__assert_not_frozen('write')
+        self.__check_capacity(len(string))
+        assert isinstance(string, hbytes)
+        original = self.index
+        self.__write(string)
+        self.forced_indices.update(hrange(original, self.index))
+        return string
+
+    def __check_capacity(self, n):
         if self.index + n > self.max_length:
             self.overdraw = self.index + n - self.max_length
             self.status = Status.OVERRUN
             self.freeze()
             raise StopTest(self.testcounter)
-        result = self._draw_bytes(self, n, distribution)
+
+    def __write(self, result):
+        initial = self.index
+        n = len(result)
         self.block_starts.setdefault(n, []).append(initial)
         self.blocks.append((initial, initial + n))
         assert len(result) == n
         assert self.index == initial
         self.buffer.extend(result)
         self.intervals.append((initial, self.index))
+
+    def draw_bytes_internal(self, n, distribution):
+        self.__assert_not_frozen('draw_bytes')
+        if n == 0:
+            return hbytes(b'')
+        self.__check_capacity(n)
+        result = self._draw_bytes(self, n, distribution)
+        assert len(result) == n
+        self.__write(result)
         return hbytes(result)
 
     def mark_interesting(self):
